@@ -8,23 +8,24 @@
 //! component scaffolding in order to transfer data back and forth across the C-style FFI layer,
 //! as well as some utilities for testing the generated bindings.
 //!
-//! The key concept here is the [`ViaFfi`] trait, which must be implemented for any type that can
-//! be passed across the FFI, and which determines:
+//! The key concept here is the [`FfiConverter`] trait, which is responsible for converting between
+//! a Rust type and a low-level C-style type that can be passed across the FFI:
 //!
-//!  * How to [represent](ViaFfi::Value) values of that type in the low-level C-style type
+//!  * How to [represent](FfiConverter::FfiType) values of the Rust type in the low-level C-style type
 //!    system of the FFI layer.
-//!  * How to ["lower"](ViaFfi::lower) rust values of that type into an appropriate low-level
+//!  * How to ["lower"](FfiConverter::lower) values of the Rust type into an appropriate low-level
 //!    FFI value.
-//!  * How to ["lift"](ViaFfi::lift) low-level FFI values back into rust values of that type.
-//!  * How to [write](ViaFfi::write) rust values of that type into a buffer, for cases
+//!  * How to ["lift"](FfiConverter::try_lift) low-level FFI values back into values of the Rust
+//!    type.
+//!  * How to [write](FfiConverter::write) values of the Rust type into a buffer, for cases
 //!    where they are part of a compound data structure that is serialized for transfer.
-//!  * How to [read](ViaFfi::read) rust values of that type from buffer, for cases
+//!  * How to [read](FfiConverter::try_read) values of the Rust type from buffer, for cases
 //!    where they are received as part of a compound data structure that was serialized for transfer.
 //!
-//! This logic encapsulates the rust-side handling of data transfer. Each foreign-language binding
+//! This logic encapsulates the Rust-side handling of data transfer. Each foreign-language binding
 //! must also implement a matching set of data-handling rules for each data type.
 //!
-//! In addition to the core` ViaFfi` trait, we provide a handful of struct definitions useful
+//! In addition to the core `FfiConverter` trait, we provide a handful of struct definitions useful
 //! for passing core rust types over the FFI, such as [`RustBuffer`].
 
 use anyhow::{bail, Result};
@@ -48,11 +49,12 @@ pub mod testing;
 pub mod deps {
     pub use anyhow;
     pub use bytes;
-    pub use ffi_support;
     pub use log;
     pub use static_assertions;
 }
 
+mod panichook;
+
 const PACKAGE_VERSION: &str = env!("CARGO_PKG_VERSION");
 
 // For the significance of this magic number 10 here, and the reason that
@@ -124,11 +126,11 @@ macro_rules! assert_compatible_version {
 
 /// Trait defining how to transfer values via the FFI layer.
 ///
-/// The `ViaFfi` trait defines how to pass values of a particular type back-and-forth over
+/// The `FfiConverter` trait defines how to pass values of a particular type back-and-forth over
 /// the uniffi generated FFI layer, both as standalone argument or return values, and as
 /// part of serialized compound data structures.
 ///
-/// (This trait is Like the `InfoFfi` trait from `ffi_support`, but local to this crate
+/// (This trait is like the `IntoFfi` trait from `ffi_support`, but local to this crate
 /// so that we can add some alternative implementations for different builtin types,
 /// and so that we can add support for receiving as well as returning).
 ///
@@ -142,7 +144,14 @@ macro_rules! assert_compatible_version {
 /// In general, you should not need to implement this trait by hand, and should instead rely on
 /// implementations generated from your component UDL via the `uniffi-bindgen scaffolding` command.
 
-pub unsafe trait ViaFfi: Sized {
+pub unsafe trait FfiConverter: Sized {
+    /// The type used in Rust code.
+    ///
+    /// For primitive / standard types, we implement `FfiConverter` on the type itself with `RustType=Self`.
+    /// For user-defined types we create a unit struct and implement it there.  This sidesteps
+    /// Rust's orphan rules (ADR-0006).
+    type RustType;
+
     /// The low-level type used for passing values of this type over the FFI.
     ///
     /// This must be a C-compatible type (e.g. a numeric primitive, a `#[repr(C)]` struct) into
@@ -160,9 +169,9 @@ pub unsafe trait ViaFfi: Sized {
     /// by (hopefully cheaply!) converting it into someting that can be passed over the FFI
     /// and reconstructed on the other side.
     ///
-    /// Note that this method takes an owned `self`; this allows it to transfer ownership
+    /// Note that this method takes an owned `Self::RustType`; this allows it to transfer ownership
     /// in turn to the foreign language code, e.g. by boxing the value and passing a pointer.
-    fn lower(self) -> Self::FfiType;
+    fn lower(obj: Self::RustType) -> Self::FfiType;
 
     /// Lift a rust value of the target type, from an FFI value of type Self::FfiType.
     ///
@@ -172,14 +181,17 @@ pub unsafe trait ViaFfi: Sized {
     ///
     /// Since we cannot statically guarantee that the foreign-language code will send valid
     /// values of type Self::FfiType, this method is fallible.
-    fn try_lift(v: Self::FfiType) -> Result<Self>;
+    fn try_lift(v: Self::FfiType) -> Result<Self::RustType>;
 
     /// Write a rust value into a buffer, to send over the FFI in serialized form.
     ///
     /// This trait method can be used for sending data from rust to the foreign language code,
     /// in cases where we're not able to use a special-purpose FFI type and must fall back to
     /// sending serialized bytes.
-    fn write<B: BufMut>(&self, buf: &mut B);
+    ///
+    /// Note that this method takes an owned `Self::RustType` because it's transfering ownership
+    /// to the foreign language code via the RustBuffer.
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>);
 
     /// Read a rust value from a buffer, received over the FFI in serialized form.
     ///
@@ -189,33 +201,11 @@ pub unsafe trait ViaFfi: Sized {
     ///
     /// Since we cannot statically guarantee that the foreign-language code will send valid
     /// serialized bytes for the target type, this method is fallible.
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self>;
-}
-
-/// A helper function to lower a type by serializing it into a buffer.
-///
-/// For complex types were it's too fiddly or too unsafe to convert them into a special-purpose
-/// C-compatible value, you can use this helper function to implement `lower()` in terms of `write()`
-/// and pass the value as a serialized buffer of bytes.
-pub fn lower_into_buffer<T: ViaFfi>(value: T) -> RustBuffer {
-    let mut buf = Vec::new();
-    ViaFfi::write(&value, &mut buf);
-    RustBuffer::from_vec(buf)
-}
-
-/// A helper function to lift a type by deserializing it from a buffer.
-///
-/// For complex types were it's too fiddly or too unsafe to convert them into a special-purpose
-/// C-compatible value, you can use this helper function to implement `lift()` in terms of `read()`
-/// and receive the value as a serialzied byte buffer.
-pub fn try_lift_from_buffer<T: ViaFfi>(buf: RustBuffer) -> Result<T> {
-    let vec = buf.destroy_into_vec();
-    let mut buf = vec.as_slice();
-    let value = <T as ViaFfi>::try_read(&mut buf)?;
-    if buf.remaining() != 0 {
-        bail!("junk data left in buffer after lifting")
-    }
-    Ok(value)
+    ///
+    /// Note the slightly unusual type here - we want a mutable reference to a slice of bytes,
+    /// because we want to be able to advance the start of the slice after reading an item
+    /// from it (but will not mutate the actual contents of the slice).
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType>;
 }
 
 /// A helper function to ensure we don't read past the end of a buffer.
@@ -223,7 +213,7 @@ pub fn try_lift_from_buffer<T: ViaFfi>(buf: RustBuffer) -> Result<T> {
 /// Rust won't actually let us read past the end of a buffer, but the `Buf` trait does not support
 /// returning an explicit error in this case, and will instead panic. This is a look-before-you-leap
 /// helper function to instead return an explicit error, to help with debugging.
-pub fn check_remaining<B: Buf>(buf: &B, num_bytes: usize) -> Result<()> {
+pub fn check_remaining(buf: &[u8], num_bytes: usize) -> Result<()> {
     if buf.remaining() < num_bytes {
         bail!(format!(
             "not enough bytes remaining in buffer ({} < {})",
@@ -234,7 +224,7 @@ pub fn check_remaining<B: Buf>(buf: &B, num_bytes: usize) -> Result<()> {
     Ok(())
 }
 
-/// Blanket implementation of ViaFfi for numeric primitives.
+/// Blanket implementation of `FfiConverter` for numeric primitives.
 ///
 /// Numeric primitives have a straightforward mapping into C-compatible numeric types,
 /// sice they are themselves a C-compatible numeric type!
@@ -243,22 +233,23 @@ macro_rules! impl_via_ffi_for_num_primitive {
     ($($T:ty),*) => {
             $(
                 paste! {
-                    unsafe impl ViaFfi for $T {
+                    unsafe impl FfiConverter for $T {
+                        type RustType = Self;
                         type FfiType = Self;
 
-                        fn lower(self) -> Self::FfiType {
-                            self
+                        fn lower(obj: Self::RustType) -> Self::FfiType {
+                            obj
                         }
 
                         fn try_lift(v: Self::FfiType) -> Result<Self> {
                             Ok(v)
                         }
 
-                        fn write<B: BufMut>(&self, buf: &mut B) {
-                            buf.[<put_ $T>](*self);
+                        fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
+                            buf.[<put_ $T>](obj);
                         }
 
-                        fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+                        fn try_read(buf: &mut &[u8]) -> Result<Self> {
                             check_remaining(buf, std::mem::size_of::<$T>())?;
                             Ok(buf.[<get_ $T>]())
                         }
@@ -276,18 +267,19 @@ impl_via_ffi_for_num_primitive! {
 ///
 /// Booleans are passed as an `i8` in order to avoid problems with handling
 /// C-compatible boolean values on JVM-based languages.
-unsafe impl ViaFfi for bool {
+unsafe impl FfiConverter for bool {
+    type RustType = Self;
     type FfiType = i8;
 
-    fn lower(self) -> Self::FfiType {
-        if self {
+    fn lower(obj: Self::RustType) -> Self::FfiType {
+        if obj {
             1
         } else {
             0
         }
     }
 
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
+    fn try_lift(v: Self::FfiType) -> Result<Self::RustType> {
         Ok(match v {
             0 => false,
             1 => true,
@@ -295,19 +287,19 @@ unsafe impl ViaFfi for bool {
         })
     }
 
-    fn write<B: BufMut>(&self, buf: &mut B) {
-        buf.put_i8(ViaFfi::lower(*self));
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
+        buf.put_i8(<bool as FfiConverter>::lower(obj));
     }
 
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         check_remaining(buf, 1)?;
-        ViaFfi::try_lift(buf.get_i8())
+        <bool as FfiConverter>::try_lift(buf.get_i8())
     }
 }
 
 /// Support for passing Strings via the FFI.
 ///
-/// Unlike many other implementations of `ViaFfi`, this passes a struct containing
+/// Unlike many other implementations of `FfiConverter`, this passes a struct containing
 /// a raw pointer rather than copying the data from one side to the other. This is a
 /// safety hazard, but turns out to be pretty nice for useability. This struct
 /// *must* be a valid `RustBuffer` and it *must* contain valid utf-8 data (in other
@@ -316,19 +308,20 @@ unsafe impl ViaFfi for bool {
 /// When serialized in a buffer, strings are represented as a i32 byte length
 /// followed by utf8-encoded bytes. (It's a signed integer because unsigned types are
 /// currently experimental in Kotlin).
-unsafe impl ViaFfi for String {
+unsafe impl FfiConverter for String {
+    type RustType = Self;
     type FfiType = RustBuffer;
 
     // This returns a struct with a raw pointer to the underlying bytes, so it's very
     // important that it consume ownership of the String, which is relinquished to the
     // foreign language code (and can be restored by it passing the pointer back).
-    fn lower(self) -> Self::FfiType {
-        RustBuffer::from_vec(self.into_bytes())
+    fn lower(obj: Self::RustType) -> Self::FfiType {
+        RustBuffer::from_vec(obj.into_bytes())
     }
 
     // The argument here *must* be a uniquely-owned `RustBuffer` previously obtained
     // from `lower` above, and hence must be the bytes of a valid rust string.
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
+    fn try_lift(v: Self::FfiType) -> Result<Self::RustType> {
         let v = v.destroy_into_vec();
         // This turns the buffer back into a `String` without copying the data
         // and without re-checking it for validity of the utf8. If the `RustBuffer`
@@ -338,18 +331,21 @@ unsafe impl ViaFfi for String {
         Ok(unsafe { String::from_utf8_unchecked(v) })
     }
 
-    fn write<B: BufMut>(&self, buf: &mut B) {
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
         // N.B. `len()` gives us the length in bytes, not in chars or graphemes.
         // TODO: it would be nice not to panic here.
-        let len = i32::try_from(self.len()).unwrap();
+        let len = i32::try_from(obj.len()).unwrap();
         buf.put_i32(len); // We limit strings to u32::MAX bytes
-        buf.put(self.as_bytes());
+        buf.put(obj.as_bytes());
     }
 
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         check_remaining(buf, 4)?;
         let len = usize::try_from(buf.get_i32())?;
         check_remaining(buf, len)?;
+        // N.B: In the general case `Buf::chunk()` may return partial data.
+        // But in the specific case of `<&[u8] as Buf>` it returns the full slice,
+        // so there is no risk of having less than `len` bytes available here.
         let bytes = &buf.chunk()[..len];
         let res = String::from_utf8(bytes.to_vec())?;
         buf.advance(len);
@@ -357,6 +353,46 @@ unsafe impl ViaFfi for String {
     }
 }
 
+/// A helper trait to implement lowering/lifting using a `RustBuffer`
+///
+/// For complex types where it's too fiddly or too unsafe to convert them into a special-purpose
+/// C-compatible value, you can use this trait to implement `lower()` in terms of `write()` and
+/// `lift` in terms of `read()`.
+pub trait RustBufferFfiConverter: Sized {
+    type RustType;
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>);
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType>;
+}
+
+unsafe impl<T: RustBufferFfiConverter> FfiConverter for T {
+    type RustType = T::RustType;
+    type FfiType = RustBuffer;
+
+    fn lower(obj: Self::RustType) -> RustBuffer {
+        let mut buf = Vec::new();
+        <T as RustBufferFfiConverter>::write(obj, &mut buf);
+        RustBuffer::from_vec(buf)
+    }
+
+    fn try_lift(v: RustBuffer) -> Result<Self::RustType> {
+        let vec = v.destroy_into_vec();
+        let mut buf = vec.as_slice();
+        let value = T::try_read(&mut buf)?;
+        if buf.remaining() != 0 {
+            bail!("junk data left in buffer after lifting")
+        }
+        Ok(value)
+    }
+
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
+        T::write(obj, buf)
+    }
+
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
+        T::try_read(buf)
+    }
+}
+
 /// Support for passing timestamp values via the FFI.
 ///
 /// Timestamps values are currently always passed by serializing to a buffer.
@@ -371,20 +407,12 @@ unsafe impl ViaFfi for String {
 /// the sign of the seconds portion represents the direction of the offset
 /// overall. The sign of the seconds portion can then be used to determine
 /// if the total offset should be added to or subtracted from the unix epoch.
-unsafe impl ViaFfi for SystemTime {
-    type FfiType = RustBuffer;
+impl RustBufferFfiConverter for SystemTime {
+    type RustType = Self;
 
-    fn lower(self) -> Self::FfiType {
-        lower_into_buffer(self)
-    }
-
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
-        try_lift_from_buffer(v)
-    }
-
-    fn write<B: BufMut>(&self, buf: &mut B) {
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
         let mut sign = 1;
-        let epoch_offset = self
+        let epoch_offset = obj
             .duration_since(SystemTime::UNIX_EPOCH)
             .unwrap_or_else(|error| {
                 sign = -1;
@@ -399,7 +427,7 @@ unsafe impl ViaFfi for SystemTime {
         buf.put_u32(epoch_offset.subsec_nanos());
     }
 
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         check_remaining(buf, 12)?;
         let seconds = buf.get_i64();
         let nanos = buf.get_u32();
@@ -421,23 +449,15 @@ unsafe impl ViaFfi for SystemTime {
 /// magnitude in seconds, and a u32 that indicates the nanosecond portion
 /// of the magnitude. The nanosecond portion is expected to be between 0
 /// and 999,999,999.
-unsafe impl ViaFfi for Duration {
-    type FfiType = RustBuffer;
+impl RustBufferFfiConverter for Duration {
+    type RustType = Self;
 
-    fn lower(self) -> Self::FfiType {
-        lower_into_buffer(self)
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
+        buf.put_u64(obj.as_secs());
+        buf.put_u32(obj.subsec_nanos());
     }
 
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
-        try_lift_from_buffer(v)
-    }
-
-    fn write<B: BufMut>(&self, buf: &mut B) {
-        buf.put_u64(self.as_secs());
-        buf.put_u32(self.subsec_nanos());
-    }
-
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         check_remaining(buf, 12)?;
         Ok(Duration::new(buf.get_u64(), buf.get_u32()))
     }
@@ -452,32 +472,24 @@ unsafe impl ViaFfi for Duration {
 /// In future we could do the same optimization as rust uses internally, where the
 /// `None` option is represented as a null pointer and the `Some` as a valid pointer,
 /// but that seems more fiddly and less safe in the short term, so it can wait.
-unsafe impl<T: ViaFfi> ViaFfi for Option<T> {
-    type FfiType = RustBuffer;
+impl<T: FfiConverter> RustBufferFfiConverter for Option<T> {
+    type RustType = Option<T::RustType>;
 
-    fn lower(self) -> Self::FfiType {
-        lower_into_buffer(self)
-    }
-
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
-        try_lift_from_buffer(v)
-    }
-
-    fn write<B: BufMut>(&self, buf: &mut B) {
-        match self {
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
+        match obj {
             None => buf.put_i8(0),
             Some(v) => {
                 buf.put_i8(1);
-                ViaFfi::write(v, buf);
+                <T as FfiConverter>::write(v, buf);
             }
         }
     }
 
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         check_remaining(buf, 1)?;
         Ok(match buf.get_i8() {
             0 => None,
-            1 => Some(<T as ViaFfi>::try_read(buf)?),
+            1 => Some(<T as FfiConverter>::try_read(buf)?),
             _ => bail!("unexpected tag byte for Option"),
         })
     }
@@ -492,32 +504,24 @@ unsafe impl<T: ViaFfi> ViaFfi for Option<T> {
 /// Ideally we would pass `Vec<u8>` directly as a `RustBuffer` rather
 /// than serializing, and perhaps even pass other vector types using a
 /// similar struct. But that's for future work.
-unsafe impl<T: ViaFfi> ViaFfi for Vec<T> {
-    type FfiType = RustBuffer;
+impl<T: FfiConverter> RustBufferFfiConverter for Vec<T> {
+    type RustType = Vec<T::RustType>;
 
-    fn lower(self) -> Self::FfiType {
-        lower_into_buffer(self)
-    }
-
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
-        try_lift_from_buffer(v)
-    }
-
-    fn write<B: BufMut>(&self, buf: &mut B) {
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
         // TODO: would be nice not to panic here :-/
-        let len = i32::try_from(self.len()).unwrap();
+        let len = i32::try_from(obj.len()).unwrap();
         buf.put_i32(len); // We limit arrays to i32::MAX items
-        for item in self.iter() {
-            ViaFfi::write(item, buf);
+        for item in obj.into_iter() {
+            <T as FfiConverter>::write(item, buf);
         }
     }
 
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         check_remaining(buf, 4)?;
         let len = usize::try_from(buf.get_i32())?;
         let mut vec = Vec::with_capacity(len);
         for _ in 0..len {
-            vec.push(<T as ViaFfi>::try_read(buf)?)
+            vec.push(<T as FfiConverter>::try_read(buf)?)
         }
         Ok(vec)
     }
@@ -531,34 +535,26 @@ unsafe impl<T: ViaFfi> ViaFfi for Vec<T> {
 /// We write a `i32` entries count followed by each entry (string
 /// key followed by the value) in turn.
 /// (It's a signed type due to limits of the JVM).
-unsafe impl<V: ViaFfi> ViaFfi for HashMap<String, V> {
-    type FfiType = RustBuffer;
+impl<V: FfiConverter> RustBufferFfiConverter for HashMap<String, V> {
+    type RustType = HashMap<String, V::RustType>;
 
-    fn lower(self) -> Self::FfiType {
-        lower_into_buffer(self)
-    }
-
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
-        try_lift_from_buffer(v)
-    }
-
-    fn write<B: BufMut>(&self, buf: &mut B) {
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
         // TODO: would be nice not to panic here :-/
-        let len = i32::try_from(self.len()).unwrap();
+        let len = i32::try_from(obj.len()).unwrap();
         buf.put_i32(len); // We limit HashMaps to i32::MAX entries
-        for (key, value) in self.iter() {
-            ViaFfi::write(key, buf);
-            ViaFfi::write(value, buf);
+        for (key, value) in obj.into_iter() {
+            <String as FfiConverter>::write(key, buf);
+            <V as FfiConverter>::write(value, buf);
         }
     }
 
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         check_remaining(buf, 4)?;
         let len = usize::try_from(buf.get_i32())?;
         let mut map = HashMap::with_capacity(len);
         for _ in 0..len {
             let key = String::try_read(buf)?;
-            let value = <V as ViaFfi>::try_read(buf)?;
+            let value = <V as FfiConverter>::try_read(buf)?;
             map.insert(key, value);
         }
         Ok(map)
@@ -570,7 +566,8 @@ unsafe impl<V: ViaFfi> ViaFfi for HashMap<String, V> {
 /// To avoid dealing with complex lifetime semantics over the FFI, any data passed
 /// by reference must be encapsulated in an `Arc`, and must be safe to share
 /// across threads.
-unsafe impl<T: Sync + Send> ViaFfi for std::sync::Arc<T> {
+unsafe impl<T: Sync + Send> FfiConverter for std::sync::Arc<T> {
+    type RustType = Self;
     // Don't use a pointer to <T> as that requires a `pub <T>`
     type FfiType = *const std::os::raw::c_void;
 
@@ -583,8 +580,8 @@ unsafe impl<T: Sync + Send> ViaFfi for std::sync::Arc<T> {
     /// Safety: when freeing the resulting pointer, the foreign-language code must
     /// call the destructor function specific to the type `T`. Calling the destructor
     /// function for other types may lead to undefined behaviour.
-    fn lower(self) -> Self::FfiType {
-        std::sync::Arc::into_raw(self) as Self::FfiType
+    fn lower(obj: Self::RustType) -> Self::FfiType {
+        std::sync::Arc::into_raw(obj) as Self::FfiType
     }
 
     /// When lifting, we receive a "borrow" of the `Arc<T>` that is owned by
@@ -592,7 +589,7 @@ unsafe impl<T: Sync + Send> ViaFfi for std::sync::Arc<T> {
     ///
     /// Safety: the provided value must be a pointer previously obtained by calling
     /// the `lower()` or `write()` method of this impl.
-    fn try_lift(v: Self::FfiType) -> Result<Self> {
+    fn try_lift(v: Self::FfiType) -> Result<Self::RustType> {
         let v = v as *const T;
         // We musn't drop the `Arc<T>` that is owned by the foreign-language code.
         let foreign_arc = std::mem::ManuallyDrop::new(unsafe { Self::from_raw(v) });
@@ -608,10 +605,9 @@ unsafe impl<T: Sync + Send> ViaFfi for std::sync::Arc<T> {
     /// Safety: when freeing the resulting pointer, the foreign-language code must
     /// call the destructor function specific to the type `T`. Calling the destructor
     /// function for other types may lead to undefined behaviour.
-    fn write<B: BufMut>(&self, buf: &mut B) {
+    fn write(obj: Self::RustType, buf: &mut Vec<u8>) {
         static_assertions::const_assert!(std::mem::size_of::<*const std::ffi::c_void>() <= 8);
-        let ptr = std::sync::Arc::clone(self).lower();
-        buf.put_u64(ptr as u64);
+        buf.put_u64(Self::lower(obj) as u64);
     }
 
     /// When reading as a field of a complex structure, we receive a "borrow" of the `Arc<T>`
@@ -619,7 +615,7 @@ unsafe impl<T: Sync + Send> ViaFfi for std::sync::Arc<T> {
     ///
     /// Safety: the buffer must contain a pointer previously obtained by calling
     /// the `lower()` or `write()` method of this impl.
-    fn try_read<B: Buf>(buf: &mut B) -> Result<Self> {
+    fn try_read(buf: &mut &[u8]) -> Result<Self::RustType> {
         static_assertions::const_assert!(std::mem::size_of::<*const std::ffi::c_void>() <= 8);
         check_remaining(buf, 8)?;
         Self::try_lift(buf.get_u64() as Self::FfiType)
@@ -639,14 +635,14 @@ mod test {
     #[test]
     fn timestamp_roundtrip_post_epoch() {
         let expected = SystemTime::UNIX_EPOCH + Duration::new(100, 100);
-        let result = SystemTime::try_lift(expected.lower()).expect("Failed to lift!");
+        let result = SystemTime::try_lift(SystemTime::lower(expected)).expect("Failed to lift!");
         assert_eq!(expected, result)
     }
 
     #[test]
     fn timestamp_roundtrip_pre_epoch() {
         let expected = SystemTime::UNIX_EPOCH - Duration::new(100, 100);
-        let result = SystemTime::try_lift(expected.lower()).expect("Failed to lift!");
+        let result = SystemTime::try_lift(SystemTime::lower(expected)).expect("Failed to lift!");
         assert_eq!(
             expected, result,
             "Expected results after lowering and lifting to be equal"
